"""Module for recording and saving measurements.
"""

import re
import threading
from collections import defaultdict, Counter


RE_FILTER_ALL_CHARACTERS = '.*'


class ProfileRecord(object):
    """Simple record to track how many times a function/path is called.
    """

    def __init__(self, name, hits):
        """Initializer.

        :param name:   String name of function or stack path.

        :param hits:   Number of times it is called.
        """
        self.name = name
        self.hits = hits

    def to_str(self):
        """Return string reprsentation."""
        result = '%s(%s=%s, %s=%s)' % (
            self.__class__.__name__, 'name', self.name, 'hits', self.hits)

        return result

    def __repr__(self):
        return self.to_str()


class CountingRecorder(object):
    """Recorder which just counts how many times something is called.
    """

    def __init__(self):
        self.db_lock = threading.Lock()
        with self.db_lock:
            self.my_db = defaultdict(lambda: 0)

    def record(self, measurement):
        """Record a measurement.

        :param measurement:     An ox_profile.core.metrics.Measurement
                                for profiling the program.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  This method is called to record a measurement. Different
                  recorders may track different things about a measurement.

        """
        with self.db_lock:
            self.my_db[measurement.name] += 1

    def query(self, re_filter=RE_FILTER_ALL_CHARACTERS, max_records=10):
        """Query the database of measurements.

        :param re_filter='.*':      String regular expression for records
                                    to include in query.

        :param max_records=10:      Maximum number of records to include.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:   The pair (result, count) where count num_records is the
                   total number of records in the database and result is
                   a list of ProfileRecord instances sorted to start from the
                   record with the most hits to the least with at most
                   max_records included.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Query the database. One slightly tricky thing to keep
                   in mind here is that name of a record is the semi-colon
                   separated list of functions for the backtrace of a
                   function. To make sure to count each function call,
                   we take apart the name into the backtrace and record
                   a hit for everything in the backtrace.

        """
        if re_filter in (None, RE_FILTER_ALL_CHARACTERS):
            regexp = None
        else:
            regexp = re.compile(re_filter)
        calls_counter = Counter()
        # Lock so we don't mess with db during query.
        # *IMPORTANT: be careful in code below to not do anything to
        # call self.record or anything else which would try to acquire
        # self.db_lock otherwise you will deadlock
        with self.db_lock:
            num_records = len(self.my_db)
            # Use explicit list in case dict changes during iteration
            for name, item in list(self.my_db.items()):
                name_list = name.split(';')
                for fname in name_list:
                    if not regexp or regexp.search(fname):
                        calls_counter[fname] += item
            my_hits = calls_counter.most_common(max_records)
            result = [ProfileRecord(name, hits) for name, hits in my_hits]
            return result, num_records

    def show(self, limit=10, query=None, sep='-', col='|'):
        """Show query as pretty formatted string.

        :arg limit=10:     Maximum lines to show.

        :arg query=None:   Optional output of calling self.query()[0]. If
                           query is None, we call self.query(max_records=100)
                           to get the query.

        :arg sep='-':      Optional horizontal line separator. Use '' if
                           you do not want a separator.

        :arg col='|':      Optional vertical line separator. Use '' if
                           you do not want a separator.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:  A string with profiling results formatted nicely.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Show the profiling results.

        """
        if query is None:
            query, num_records = self.query(max_records=100)
        else:
            num_records = len(query)
        total_hits = float(sum([item.hits for item in query]))
        if limit:
            note = '' if limit is None else (' (top %i/%i results)' % (
                min(limit, num_records), num_records))
            query = query[:limit]
        else:
            note = ''
        width = 40
        fmt = '%s {:%i} %s {:>8} %s {:>5} %s' % (col, width, col, col, col)
        header = fmt.format('Function', 'Hits', '%')
        if sep:
            line_sep = '\n  ' + (sep * len(header)) + '\n  '
        else:
            line_sep = '\n'
        text = ('Profiling results:%s\n%s' % (note, line_sep)) + (
            header + line_sep) + line_sep.join([fmt.format(
                i.name[:width], i.hits, '%.1f' % (100*(i.hits/total_hits)))
                                                  for i in query]) + line_sep

        return text



"""Module for handling measurement and sampling of data.
"""


class Measurement(object):
    """Measurement of profiling information.

    The idea is that you initialize an instance of a Measurement when
    you want to get some profiling information about a program.

    One important thing the Measurement instance must do is take
    a stack frame and map it to a unique name using the `snap` method.
    The name of the measurement is basically the full stack backtrace.
    Thus measurements with the same stack backtrace will have the same
    name (so we can count how often they occur).

    Note that the `query` method in the `CountingRecorder` class takes
    apart the name into the list of functions for the backtrace and
    records a hit for each of those. See the `query` method for details.
    """

    def __init__(self, frame):
        """Initializer.

        :param frame:    Stack frame to measure.

        """
        self.name = self.snap(frame)

    def get_path(self):
        """Return backtrace path for snapped measurement.
        """
        return self.name.split(';')

    def snap(self, frame):
        """Snap a measurement for the given stack frame (called by __init__).

        :param frame:     Stack frame to take a measurement about.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:   String representing nicely formatted backtrace of functions
                   for given stack frame.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Meant to be called internally by __init__ to snap a
                   measurement. We want each call path to have a different
                   string representation. To do that we just use the
                   backtrace as the name.

        """
        stack = []
        self.name = '%s(%s)' % (frame.f_code.co_name,
                                frame.f_globals.get('__name__'))
        while frame is not None:
            formatted_frame = '%s(%s)' % (frame.f_code.co_name,
                                          frame.f_globals.get('__name__'))
            stack.append(formatted_frame)
            frame = frame.f_back

        formatted_stack = ';'.join(reversed(stack))
        return formatted_stack



"""Tools to sample running programs.
"""

import doctest
import logging
import sys

from ox_profile.core import metrics


class Freezer(object):
    """
    Muck with switch/check interval to prevent thread context switching while
    trying to capture profiling information for safety
    """
    def __init__(self):
        if sys.version_info >= (3, 2, 0):
            # https://docs.python.org/3/library/sys.html#sys.setswitchinterval
            # New in version 3.2.
            self._get_interval = sys.getswitchinterval
            self._set_interval = sys.setswitchinterval
            self._interval = 10000  # number of seconds to wait
            self._log_msg_template = "Switch interval now %.2f"
        else:
            # https://docs.python.org/3/library/sys.html#sys.setcheckinterval
            # Deprecated since version 3.2
            self._get_interval = sys.getcheckinterval
            self._set_interval = sys.setcheckinterval
            self._interval = 100000  # number of instructions to wait
            self._log_msg_template = "Check interval now %.2f"

    def __enter__(self):
        logging.debug('Process sampling')
        self._stored_interval_value = self._get_interval()
        self._set_interval(self._interval)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._set_interval(self._stored_interval_value)
        logging.debug(self._log_msg_template, self._stored_interval_value)


class Sampler(object):
    """Basic class to sample program for statistical profiling.

This is the workhorse for ox_profile to do statistical sampling.
At a low level, we do this sampling using `sys._current_frames`. As
suggested by the leading underscore, this system function may be a bit
less robust. Indeed, the documentation says "This function should be
used for specialized purposes only." Hopefully the core python
developers will not make major changes to such a useful function.

The most interesting method is `run` which:

  1. Calls `Freezer` to try and prevent a thread context switch.
  2. Calls `sys._current_frames` to sample what the python interpreter is
     doing.
  3. Updates a simple in-memory database of what functions are running.

In principle, you could just use the Sampler via something like

>>> from ox_profile.core import sampling, recording
>>> sampler = sampling.Sampler(recording.CountingRecorder())
>>> def foo():
...     sampler.run()
...     return 'done'
...
>>> dummy = foo()
>>> info, calls = sampler.my_db.query()
>>> pretty = ('%c'%10).join(map(str, info))
>>> print('%i calls:%c%s' % (calls, 10, pretty)) # doctest: +ELLIPSIS
1 calls:
ProfileRecord(...)

The above would have the sampler take a snapshot of the stack frames
when the `foo` function is run. Of course, this isn't very useful by
itself because it just tells you that `foo` is being run. It could be
useful if there were other threads which were running because the
sampler would tell you what stack frame those threads were in. For
this reason you should use the SimpleLauncher class in
the `ox_profile/core/launchers` module which creates an instance
of this class.

    """

    def __init__(self, my_db, freezer=None):
        self.my_db = my_db
        self.freezer = freezer or Freezer()

    def show(self, *args, **kwargs):
        """Syntactic sugar self.my_db.show(*args, **kwargs) to show results.

        Returns a string describing collected data.
        """
        return self.my_db.show(*args, **kwargs)

    def query(self, *args, **kwargs):
        """Syntactic sugar self.my_db.query(*args, **kwargs) to query results.

        Returns a string describing collected data.
        """
        return self.my_db.query(*args, **kwargs)

    def get_measure_tool(self):
        """Get a class or function to call to take a measurement.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  class or function to call to take a measurement.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  This method provides the "measurement tool" we are going
                  to use in profiling. Sub-classes could override this to
                  take different kinds of measurements.

        """
        return metrics.Measurement

    def run(self):
        """Run the sampler to make a measurement of the current stack frames.
        """
        measure_tool = self.get_measure_tool()

        with self.freezer:
            for dummy_frame_id, frame in (
                    sys._current_frames(  # pylint: disable=protected-access
                        ).items()):
                self.my_db.record(measure_tool(frame))

    def __call__(self, *args, **kwargs):
        """Syntactic sugar to call `self.run(*args, **kwargs)`."""

        return self.run(*args, **kwargs)


if __name__ == '__main__':
    # Run doctest if file executed as a script
    doctest.testmod()
    print('Finished Tests')



"""Basic views for ox_profile flask UI.
"""

import urllib.parse
import io
import csv
import collections
import datetime
import logging
from functools import wraps

from flask import (
    request, Markup, current_app, render_template, url_for, g, make_response)
from flask_login import login_required
from flask_login import current_user

from ox_profile.ui.flask import OX_PROF_BP, ReqRecord

RouteInfo = collections.namedtuple('RouteInfo', [
    'name', 'hits', 'avg_time'])

def access_problem_p():
    """If user is allowed access to ox_profile, return None else return error.

    This is meant to be called when a user tries to access ox_profile routes
    so that non-privledged users cannot cause trouble. It looks in
    current_app.config['OX_PROF_USERS'] for a dictionary or set of allowed
    users.
    """
    allowed = current_app.config.get('OX_PROF_USERS', {})
    name_field = current_app.config.get(
        'OX_PROF_USERNAME_FIELD', 'name')
    effective_username = getattr(current_user, name_field)
    if effective_username in allowed:
        return None

    msg = 'Current user.%s = %s not in the %i allowed ox profile users.%s' % (
        effective_username, name_field, len(allowed),
        '\nConfigure app.config["OX_PROF_USERS"] if necessary.')
    return render_template('ox_prof_err.html', error_msg=msg)


def restrict_access(my_func):
    """Simple decorator to call access_problem_p before execution.
    """
    @wraps(my_func)
    def wrapper(*args, **kwds):
        """Check access_problem_p and then execute wrapped function.
        """

        problem = access_problem_p()
        if problem is None:
            return my_func(*args, **kwds)
        return problem

    return wrapper


@OX_PROF_BP.route('/')
@OX_PROF_BP.route('/index')
@login_required
@restrict_access
def index():
    """Top-level index for ox profile.
    """
    commands = [(n, url_for('%s.%s' % ('ox_profile', n))) for n in [
        'status', 'pause', 'unpause', 'show_req_times']]

    return render_template('ox_prof_intro.html', commands=commands)


@OX_PROF_BP.route('/status')
@login_required
@restrict_access
def status():
    """Show status of current profiling.
    """
    re_filter = request.args.get('re_filter', '.*')
    max_records = int(request.args.get('max_records', 50))

    query, total_records = OX_PROF_BP.launcher.sampler.my_db.query(
        re_filter=re_filter, max_records=max_records)

    return render_template(
        'ox_prof_status.html', launcher=OX_PROF_BP.launcher,
        max_records=max_records, total_records=total_records, query=query)


@OX_PROF_BP.route('/pause')
@login_required
@restrict_access
def pause():
    """Pause the profiler.
    """
    OX_PROF_BP.launcher.pause()
    return render_template('ox_prof_msg.html', message='paused')


@OX_PROF_BP.route('/unpause')
@login_required
@restrict_access
def unpause():
    """Unpause profiler or start it for the first time.
    """
    launcher = OX_PROF_BP.launcher
    msgs = []
    if not launcher.is_alive():
        msgs.append('Started thread for first time')
        launcher.start()
    launcher.unpause()
    msgs.append('unpaused')
    return render_template('ox_prof_msg.html', message=Markup(
        '\n<BR>\n'.join(msgs)))


@OX_PROF_BP.route('/set_interval')
@login_required
@restrict_access
def set_interval():
    "Set the sampling interval"
    interval = request.args.get('interval', None)
    problem = None
    try:
        fint = float(interval)
        if fint < 0:
            raise ValueError('Negative interval of %s illogical.')
        if fint > 10:
            raise ValueError('Interval > 10 is useless.')
    except ValueError as problem:
        logging.debug('Could not convert interval %s to float because %s',
                      interval, problem)
    if problem:
        return render_template('ox_prof_err.html', error_msg=problem)
    OX_PROF_BP.launcher.set_interval(fint)

    return render_template('ox_prof_msg.html', message=(
        'Changed sampling interval to %f' % fint))


@OX_PROF_BP.before_app_request
def monitor_routes():
    """Simple function to record start time of request.

    See monitor_route_completion for more info.
    """
    g.ox_prof_ts = datetime.datetime.utcnow()


@OX_PROF_BP.teardown_app_request
def monitor_route_completion(exception=None):
    """Monitor and record route completion.

    :param exception=None:    Provided by flask if exception happened.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:   Record a few things in OX_PROF_BP about the route that
               just completed for profiling information.

    """
    try:
        user = getattr(current_user, 'name', None)
        if user is None:
            user = getattr(current_user, 'username', 'unknown')
        OX_PROF_BP.record_req(user, request.endpoint, g.ox_prof_ts,
                              datetime.datetime.utcnow())
    except Exception as problem:  # pylint:disable=broad-except
        logging.error('Got problem %s in monitor_route_completion', str(
            problem))


@OX_PROF_BP.route('/show_req_times')
@login_required
@restrict_access
def show_req_times():
    """Show statistical summary of request times for each route.
    """
    reqs = OX_PROF_BP.get_reqs()
    if int(request.args.get('as_csv', 0)):
        return _make_csv_response('reqs.csv', reqs)

    user_times = {}
    func_times = {}
    for (user, endpoint), time_list in reqs.items():
        runtimes = [(e - s).total_seconds() for s, e in time_list]
        fdata = func_times.get(endpoint, [])
        if not fdata:
            func_times[endpoint] = fdata
        fdata.extend(runtimes)
        udata = user_times.get(user, [])
        if not udata:
            user_times[endpoint] = udata
        udata.extend(runtimes)
    rinfo = [RouteInfo(name, len(data), sum(data)/float(len(data))) for (
        name, data) in func_times.items()]

    result = render_template('ox_prof_rinfo.html', rinfo=reversed(sorted(
        rinfo, key=lambda r: r.avg_time)), csv_link=Markup(make_download_link(
            request, {'as_csv': 1})))

    return result


def _make_csv_response(name, data):
    if isinstance(data, str):
        text = data
    elif isinstance(data, dict):
        csv_data = io.StringIO()
        writer = csv.writer(csv_data)
        writer.writerow(['user', 'endpoint'] + list(ReqRecord._fields))
        for (user, endpoint), req_data in data.items():
            writer.writerows([[user, endpoint] + list(i) for i in req_data])
        del writer
        csv_data.seek(0)
        text = csv_data.read()
    else:
        raise TypeError('Cannot convert type %s to CSV.' % type(data))

    response = make_response(text)
    response.headers["Content-Disposition"] = (
        "attachment; filename=%s" % name)
    return response


def make_download_link(my_request, updates, text='Download CSV'):
    """Helper function to make a download link to give request.

    :arg my_request:        A request object we are linking to.

    :arg updates:           A dict of parameters to override in my_request.

    :arg text='DownloadCSV':   Text string of download link.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :returns:   String representing HTML download link to request.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:    Make a simple download link to a request.

    """
    url_params = my_request.args.to_dict()
    url_params.update(updates)
    return '<A HREF="%s/%s?%s">%s</A>' % (
        my_request.url_root.rstrip('/'), my_request.path.lstrip('/'),
        urllib.parse.urlencode(url_params), text)



import unittest
from time import sleep

from ox_profile.core.launchers import SimpleLauncher


def one_second_running_function():
    sleep(1)


def three_seconds_running_function():
    sleep(3)


class OxProfileIntegrationTestCase(unittest.TestCase):

    def test_simple_launcher_sampler_query(self):
        launcher = SimpleLauncher(interval=.001)
        launcher.start()
        launcher.unpause()

        one_second_running_function()

        launcher.is_alive()
        launcher.cancel()
        query, total_records = launcher.sampler.my_db.query(max_records=None)
        self.assertGreater(total_records, 0)

        query_result = [(i.name, i.hits) for i in query if "one_second_running_function" in i.name]
        self.assertEqual(len(query_result), 1)
        name, hits = query_result.pop()
        self.assertGreater(hits, 0)

    def test_simple_launcher_sampler_query_calls(self):
        launcher = SimpleLauncher(interval=.001)
        launcher.start()
        launcher.unpause()

        one_second_running_function()
        three_seconds_running_function()

        launcher.is_alive()
        launcher.cancel()
        query, total_records = launcher.sampler.my_db.query(max_records=None)
        self.assertGreater(total_records, 0)

        one_second_running_function_result = next(
            ((i.name, i.hits) for i in query if "one_second_running_function" in i.name), None
        )
        three_seconds_running_function_result = next(
            ((i.name, i.hits) for i in query if "three_seconds_running_function" in i.name), None
        )

        self.assertIsNotNone(one_second_running_function_result)
        self.assertIsNotNone(three_seconds_running_function_result)
        one_second_running_function_calls = one_second_running_function_result[1]
        three_seconds_running_function_calls = three_seconds_running_function_result[1]
        self.assertGreater(three_seconds_running_function_calls, one_second_running_function_calls)


if __name__ == '__main__':
    unittest.main()
